// AppDlg.cpp : implementation file
//

#include "stdafx.h"
#include "App.h"
#include "serviceControl.h"
#include "cancel.h"
#include "winioctl.h"
#include "AppDlg.h"
#include "afxErrorMsg.h"
#include "warning.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define MAX_IOREQUESTS 1000

CString logBuffer[50];
int currentLine = 0;
ULONG logNumber=0;

const CString newLine("\r\n");
const CString marker("---");

void putText(CString& buffer, CString& string)
{
    int index;

    if (currentLine >= 50) {
        currentLine = 0;
    }

    logBuffer[currentLine++].Format("%08d %s %s", logNumber++, string, newLine);
    
    buffer.Empty();
    for (index = 0; index < 50; index++) {
        if (logBuffer[index].GetLength() == 0) {
            break;
        }
        buffer += logBuffer[index];
    }
}

void initLog()
{
    for (int index=0; index < 50; logBuffer[index++].Empty());
}


OVERLAPPED overlapped;
DWORD completedIOs = 0;
DWORD cancelledIOs = 0;

void  CALLBACK FileIOCompletionRoutine(
  DWORD dwErrorCode,                // completion code
  DWORD dwNumberOfBytesTransfered,  // number of bytes transferred
  LPOVERLAPPED lpOverlapped         // pointer to structure with I/O 
                                    // information
)
{
	//
	// the completion apc
	//
	switch(dwErrorCode) {
	case 0:
		completedIOs++;
		break;

	default:
		cancelledIOs++;
		break;
	}

}

#define Thread  __declspec( thread )

const ULONG writeChunk = 256;
Thread char workBuffer[writeChunk * MAX_IOREQUESTS];

class WorkBuffer {

    WorkBuffer(ULONG size) { buffer = new char[writeChunk * MAX_IOREQUESTS]; }
    ~WorkBuffer();
    char * bufferp() { return buffer; };

    private:
        char * buffer;
};

WorkBuffer::~WorkBuffer()
{
#ifdef _DEBUG
    for (int i = writeChunk * MAX_IOREQUESTS; 
         i > 0; 
         buffer[--i] = 0 );
#endif

    if (buffer) {

        delete[] buffer;
    }
}


UINT workRoutine( LPVOID context)
{
    THREAD_CONTEXT  threadContext = *(PTHREAD_CONTEXT) context;    
    ULONG writesIssued = 0;
    BOOL result = TRUE;
    ULONG writesCompleted = 0;
    // WorkBuffer wb;
    //
    // generate a lot of async writes to the handle 
    //
    while (result && (threadContext.ioOperations > writesIssued)) {

        overlapped.Offset = writesIssued * writeChunk;
        overlapped.OffsetHigh = 0;
        overlapped.hEvent = NULL; // we don't wait on events        
        //
        // in the checked build if you send too many writes down for the same buffer using method direct
        // the memory manager complains about the number of mappings.
        // so instead slide a window across a large enough buffer
        //
        char * windowBuf = &workBuffer[writeChunk*writesIssued];
        //
        // fill the output segment with testable values
        //
        ULONG * sequence = (ULONG *) windowBuf;
        *sequence = writesIssued;
        char * ptr = &windowBuf[sizeof ULONG];
        for (int index = 0; index < (writeChunk - (sizeof ULONG)); index++) {
            ptr[index] = index;
        }
        //
        // do the write thing :-)
        //
        result = WriteFileEx(threadContext.handle, windowBuf, writeChunk, &overlapped, FileIOCompletionRoutine);
        writesIssued++;

        if (!result) {            
            //
            // some sort of problem?
            //
            AfxMessageBox(AfxErrorMsg(CString("worker thread WriteFileEx error")));
            //
            // this is complete, sort of.
            //
            writesCompleted++;
     
        } else {
            //
            // the operation is pending
            //
            if ((writesIssued % 10) == 0) {
                //
                // every now and then take a nap
                //
                SleepEx(1, TRUE);
            }
        }
    }

    //
	// CancelIo(threadContext.handle)?
	//
    
    //
    // post a completion message and then die.
    //
    ::PostMessage(threadContext.hWnd, WM_USER_THREAD_TERMINATED, result, writesIssued - writesCompleted );    

    AfxEndThread(0);


    return (!result);

}

/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// no generated messages
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CAppDlg dialog

CAppDlg::CAppDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CAppDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CAppDlg)
	m_cancelIrp = FALSE;
	m_cancelOn = FALSE;
	m_completeOn = FALSE;
	m_speed = 1;
	m_IoOperations = 1;
	m_queuelen = 0;
	m_completeCanceled = FALSE;
	m_earlyRelease = FALSE;
	m_nThreads = 1;
	m_nPasses = 1;
	m_passesComplete = 0;
	m_iosCompleted = 0;
	m_cancelledIO = 0;
	m_statusMessages = _T("");
	m_setCancel = FALSE;
	m_apc_complete = 0;
	m_unloadDriver = FALSE;
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
    driver = NULL;
    m_userWarned = FALSE;
}

CAppDlg::~CAppDlg()
{
    if (driver) {
        delete driver;
        driver = NULL;
    }
}

void CAppDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAppDlg)
	DDX_Control(pDX, IDC_NPASSES, m_passEdit);
	DDX_Control(pDX, IDOK, m_exitButton);
	DDX_Control(pDX, IDC_SPIN_PASSES, m_spinPasses);
	DDX_Control(pDX, IDC_NTHREADS, m_threadsEdit);
	DDX_Control(pDX, IDC_IO_OPERATIONS, m_ioEdit);
	DDX_Control(pDX, IDC_SPIN_THREADS, m_threadSpin);
	DDX_Control(pDX, IDC_SPIN1, m_ioOpSpinCtrl);
	DDX_Control(pDX, IDC_START_IO, m_startIO);
	DDX_Control(pDX, IDC_SET_STATE, m_setState);
	DDX_Control(pDX, IDC_GET_STATE, m_getState);
	DDX_Control(pDX, IDC_STARTSTOP, m_startstop);
	DDX_Check(pDX, IDC_CHK_CANCELIRP, m_cancelIrp);
	DDX_Check(pDX, IDC_CHK_CANCELON, m_cancelOn);
	DDX_Check(pDX, IDC_CHK_COMPLETEON, m_completeOn);
	DDX_Radio(pDX, IDC_RADIO_SPEED, m_speed);
	DDX_Text(pDX, IDC_IO_OPERATIONS, m_IoOperations);
	DDV_MinMaxInt(pDX, m_IoOperations, 1, 1000);
	DDX_Text(pDX, IDC_QUEUELEN, m_queuelen);
	DDX_Check(pDX, IDC_CHK_COMPLETECANCELLED, m_completeCanceled);
	DDX_Check(pDX, IDC_CHK_EARLYRELEASE, m_earlyRelease);
	DDX_Text(pDX, IDC_NTHREADS, m_nThreads);
	DDV_MinMaxInt(pDX, m_nThreads, 1, 32);
	DDX_Text(pDX, IDC_NPASSES, m_nPasses);
	DDV_MinMaxUInt(pDX, m_nPasses, 1, 10000);
	DDX_Text(pDX, IDC_PASSES_COMPLETE, m_passesComplete);
	DDX_Text(pDX, IDC_NCOMPLETE, m_iosCompleted);
	DDX_Text(pDX, IDC_NCANCELLED, m_cancelledIO);
	DDX_Text(pDX, IDC_EDIT_STATUS, m_statusMessages);
	DDX_Check(pDX, IDC_CHK_SET_CANCEL, m_setCancel);
	DDX_Text(pDX, IDC_APC_COMPLETE, m_apc_complete);
	DDX_Check(pDX, IDC_CHECK_UNLOAD, m_unloadDriver);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAppDlg, CDialog)
	//{{AFX_MSG_MAP(CAppDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_BN_CLICKED(IDC_STARTSTOP, OnStartstop)
	ON_BN_CLICKED(IDC_GET_STATE, OnGetState)
	ON_BN_CLICKED(IDC_SET_STATE, OnSetState)
	ON_BN_CLICKED(IDC_START_IO, OnGenerateIo)
    ON_MESSAGE(WM_USER_THREAD_TERMINATED, OnThreadTerminate)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CAppDlg message handlers

BOOL CAppDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	
	// TODO: Add extra initialization here

	LPTSTR cwdBuf;
	DWORD bufSize;

	bufSize = GetCurrentDirectory(0, NULL);
	ASSERT(bufSize);
	cwdBuf = cwd.GetBufferSetLength(bufSize);
	bufSize = GetCurrentDirectory(bufSize, cwdBuf);
	ASSERT(bufSize && (bufSize <= (DWORD) cwd.GetLength()));
	cwd.ReleaseBuffer();

    //
    // set the spin control min max, etc
    //
    m_ioOpSpinCtrl.SetBase(10);
    m_ioOpSpinCtrl.SetRange(1, MAX_IOREQUESTS);
    m_ioOpSpinCtrl.SetPos(10);

    m_threadSpin.SetBase(10);
    m_threadSpin.SetRange(1, 32);
    m_threadSpin.SetPos(1);

    m_spinPasses.SetBase(10);
    m_spinPasses.SetRange(1, 10000);
    m_spinPasses.SetPos(1);

    initLog();

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CAppDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CAppDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CAppDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

void CAppDlg::OnStartstop() 
{
	// TODO: Add your control notification handler code here
    if (driver == NULL) 
    {
        //
        // start the driver
        //
        if (!m_userWarned) {

            CWarning dlg;
	
	        int nResponse = dlg.DoModal();
	        if (nResponse == IDOK)
	        {
		        m_userWarned = TRUE;
	        }
	        else if (nResponse == IDCANCEL)
	        {
                //
                // we won't exit until the user pushes the exit button,
                // but we will keep returing here instead.
                //
		        return;
	        }
        }

        CString path = cwd + CString("\\") + CString(DEVICE_NAME) + CString(".sys");
	    try {

		    driver = new ServiceControl(DEVICE_NAME, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_OVERLAPPED, path);
	    } 
	    
	    catch(...) {

		    AfxMessageBox(AfxErrorMsg(CString("Unexpected exception in create service control")));
            //
            // since we didn't throw this exception it is unclear what to do,
            // do nothing.
            //
            driver = NULL;
	    }

        if (driver) {
            //
            // see if everything really went ok
            //
            DWORD error = driver->getError();

            if (error != 0) {
                
		        AfxMessageBox(AfxErrorMsg(error, CString("create service control")));
            
                delete driver;
                driver = NULL;
            }

	    }

        if (driver) {
            

            //
            // change the button from Start Driver to Stop Driver
            //
            activateDriverDlg(); 
            //
            // allow the user to set thread controls
            //
            setThreadControls(TRUE);
            //
            // get the driver state
            //
            OnGetState();
        }

    } else {
        
        //
        // stop the driver
        //
        driver->UnloadDriverOnExit(m_unloadDriver);

        delete driver;
        driver = NULL;

        deactivateDriverDlg();
        //
        // turn off thread controls
        //
        setThreadControls(FALSE);
		//
		// but turn start driver back on
		//
		m_startstop.EnableWindow(TRUE);
    }

	
}


void CAppDlg::OnGetState() 
{
	// TODO: Add your control notification handler code here
    ASSERT(driver);

    DWORD outputSize = 0;
    BOOL result = 
        DeviceIoControl(driver->handle(),
                        (DWORD) IOCTL_CANCEL_GET_FLAGS,
                        NULL, (DWORD) 0, // no input data to driver
                        &cancelState, (DWORD) sizeof(cancelState), &outputSize, // output from driver
                        NULL);
    if (result == TRUE) {
        //
        // set the state of the driver
        //
        m_cancelIrp =         FLAG_ON(cancelState.flags, CANCEL_IRP);
        m_cancelOn =          FLAG_ON(cancelState.flags, CANCEL_ON);
        m_completeOn =        FLAG_ON(cancelState.flags, COMPLETE_ON);
        m_completeCanceled =  FLAG_ON(cancelState.flags, COMPLETE_CANCELED);
        m_earlyRelease =      FLAG_ON(cancelState.flags, EARLY_RELEASE);
        m_setCancel =         FLAG_ON(cancelState.flags, SET_CANCEL);

        m_speed = cancelState.speed;
        ASSERT((m_speed >= SLOW) && (m_speed <= FAST));
        
        m_queuelen = cancelState.queueLength;

        

    } else {
        //
        // display some error message
        //
        putText(m_statusMessages, CString(AfxErrorMsg(CString("IOCTL_CANCEL_GET_FLAGS failed"))));
        
    }
    UpdateData(FALSE);
	
}



void CAppDlg::OnSetState() 
{
	// TODO: Add your control notification handler code here
    UpdateData(TRUE);

    cancelState.flags = 0;
    if (m_cancelIrp) {
        cancelState.flags |= CANCEL_IRP;
    }
    if (m_cancelOn) {
        cancelState.flags |= CANCEL_ON;
    }
    if (m_completeOn) {
        cancelState.flags |= COMPLETE_ON;
    }
    if (m_completeCanceled) {
        cancelState.flags |= COMPLETE_CANCELED;
    }
    if (m_earlyRelease) {
        cancelState.flags |= EARLY_RELEASE;
    }
    if (m_setCancel) {
        cancelState.flags |= SET_CANCEL;
    }


    cancelState.speed = m_speed;
    ASSERT((cancelState.speed >= SLOW) && (cancelState.speed  <= FAST));

    DWORD outputSize = 0;

    BOOL result = 
        DeviceIoControl(driver->handle(),
                        (DWORD) IOCTL_CANCEL_SET_FLAGS,
                        &cancelState, (DWORD) sizeof(cancelState), // input to driver
                        NULL, (DWORD) 0, &outputSize, // no output from driver                        
                        NULL);
    if (!result) {
        //
        // display some error message
        //
        putText(m_statusMessages, CString(AfxErrorMsg(CString("IOCTL_CANCEL_SET_FLAGS failed")))); 
    
    } 
    //
    // see what the driver actually did
    //
    OnGetState();
	
}

void CAppDlg::startThreads()
{

    //
    // fire up the new threads
    //
    for (int i = 0; i < m_nThreads; i++) {
        workerThread = AfxBeginThread( workRoutine,  &context);
        if (workerThread == NULL) {
            break;
        }
    }
    threadsStarted = i;
    completionsExpected = threadsStarted * m_IoOperations;
    ASSERT(threadsStarted > 0);	
}


void CAppDlg::OnGenerateIo() 
{
    ASSERT(driver);


    //
    // just in case the user modified the settings, make the driver
    // operational state current with the dialogue
    //
    OnSetState();

    
    m_iosCompleted = 0;
    m_passesComplete = 0;
    completedIOs = 0;
    cancelledIOs = 0;
    m_cancelledIO = 0;
	m_apc_complete = 0;

    UpdateData(FALSE);
    
	//
    // start a worker thread to send m_IoOperations to driver->handle() and
    // then simply terminate.
    //
    context.handle = driver->handle();
    context.ioOperations = m_IoOperations;
    context.hWnd = m_hWnd;
    
    // disable the dialogue controls associate with threads
    //
    setThreadControls(FALSE);
    
    startThreads();
   
}

void CAppDlg::OnThreadTerminate(WPARAM result, LPARAM writes)
{
    CString msg;
    BOOL status = TRUE;


    ASSERT(threadsStarted > 0);
    //
    // adjust the number of outstanding IO operations
    //
    ASSERT(m_IoOperations >= writes);

    
    m_iosCompleted += writes;
    m_cancelledIO = cancelledIOs;
	m_apc_complete = completedIOs;
    //
    // get the current driver state
    //
    OnGetState(); 

    //
    // see if this is the last thread
    //
    //
    // tell the use what happened
    //
    msg.Format("Worker thread wrote %d buffers with result %d", writes, result);
    putText(m_statusMessages, msg);
    UpdateData(FALSE);

    if (--threadsStarted == 0) {

        m_passesComplete++;
        UpdateData(FALSE);

        msg.Format("Pass %d complete", m_passesComplete);
        putText(m_statusMessages, msg);
        //
        // we have finished one pass,
        //
        //
        if (m_nPasses <= m_passesComplete) {
            //
            //
            // enable the thread controls
            //
            setThreadControls(TRUE);

        } else {

            startThreads();
        }
    }
    
    UpdateData(FALSE);
    

}

void CAppDlg::setThreadControls(BOOLEAN state)
{
    if (driver && state) {

        m_startIO.EnableWindow(TRUE);
        m_exitButton.EnableWindow(TRUE);

    } else {

		m_startIO.EnableWindow(FALSE);

        if (!driver) {
            m_exitButton.EnableWindow(TRUE);
        } else {
            m_exitButton.EnableWindow(FALSE);
        }
	}

    m_threadSpin.EnableWindow(state);
    m_ioEdit.EnableWindow(state);
    m_threadsEdit.EnableWindow(state);
    m_ioOpSpinCtrl.EnableWindow(state);
    m_getState.EnableWindow(state);
    m_setState.EnableWindow(state);
    m_passEdit.EnableWindow(state);
    
    

}

void CAppDlg::activateDriverDlg()
{
    m_startstop.SetWindowText("Stop Driver");
    //
    // enable the driver operation buttons
    //
    m_getState.EnableWindow(TRUE);
    m_setState.EnableWindow(TRUE);
    m_startIO.EnableWindow(TRUE);

    
}

void CAppDlg::deactivateDriverDlg()
{
    m_startstop.SetWindowText("Start Driver");
    //
    // disable the driver operation buttons
    //
    m_getState.EnableWindow(FALSE);
    m_setState.EnableWindow(FALSE);
    m_startIO.EnableWindow(FALSE);

}


