#ifndef CANCEL_H
#define CANCEL_H

#define CANCEL_DEVICE_TYPE 60301
#define DEVICE_NAME         "Cancel"
#define NT_DEVICE_NAME      L"\\Device\\Cancel"
#define DOS_DEVICE_NAME     L"\\DosDevices\\Cancel"
//
// operational state bits for device extension state field
//
#define CANCEL_ON         0x00000001 // turns on or off doing anything in cancel routine
#define CANCEL_IRP        0x00000002 // cancel the input Irp or the first cancel'd irp in the queue?
#define COMPLETE_ON       0x00000004 // do anything in timer dpc?
#define COMPLETE_CANCELED 0x00000008 // if Irp->Cancel complete the Irp anyway?
#define EARLY_RELEASE     0x00000010 // release cancel spinlock at top of cancel routine
#define SET_CANCEL        0x00000020 // set the cancel routine on dispatch
#define VALID_FLAGS       0x0000003F

#define SLOW            0x00000000
#define MED_SLOW        0x00000001
#define MED_FAST        0x00000002
#define FAST            0x00000003

#define SLOW_SPEED      -5000000; //  2HZ
#define MED_SLOW_SPEED  -1250000; //  8HZ
#define MED_FAST_SPEED   -625000; // 16HZ
#define FAST_SPEED       -312500; // 32HZ 

#ifndef FLAG_ON
#define FLAG_ON(flag, bit) (((flag) & (bit)) ? TRUE : FALSE)
#endif

//
// IOCTL to set the device state
//
// Input buffer == CANCEL_STATE
// Output buffer (none)
//
#define IOCTL_CANCEL_SET_FLAGS CTL_CODE(CANCEL_DEVICE_TYPE, 2048, METHOD_BUFFERED, FILE_ANY_ACCESS)
//
// IOCTL to get the device state and queue length
//
// Input buffer (none)
// Output buffer == CANCEL_STATE
//
#define IOCTL_CANCEL_GET_FLAGS CTL_CODE(CANCEL_DEVICE_TYPE, 2049, METHOD_BUFFERED, FILE_ANY_ACCESS)

typedef struct {

    ULONG flags;
    ULONG speed;
    ULONG queueLength; // ignored on Set operations

} CANCEL_STATE, *PCANCEL_STATE;


#endif
